#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "defines.h"
#include "globals.h"
#include "protos.h"

/***************************************************************/
/*                                                             */
/*  DOSUBST.C                                                  */
/*                                                             */
/*  Performs line substitution for reminders.                  */
/*                                                             */
/*  If mode = 0, ignore %" escapes.                            */
/*  If mode = 1, process %" escapes.  If a RUN type reminder   */
/*     does not have %" escape, return the empty string.       */
/*                                                             */
/***************************************************************/

static char TODAY[] = "today";
static char TOMORROW[] = "tomorrow";

#ifdef __STDC__
int DoSubst(char *src, char *dst, int d, int m, int y, int jul, enum Token_t t, int tim, int mode)
#else
int DoSubst(src, dst, d, m, y, jul, t, tim, mode)
     char *src;
     char *dst;
     int d;
     int m;
     int y;
     int jul;
     enum Token_t t;
     int tim;
     int mode;
#endif
{
   int diff = jul - JulianToday;
   char c;
   char *od, *s;
   int wkday = jul % 7;
   char *plu, *pm, *cpm;
   int curtim = (int) (SystemTime() / 60);
   int done;
   int h;
   int hh;
   int min;
   int ch;
   int chh;
   int cmin;
   int tdiff;
   int adiff, mdiff, hdiff;
   char *mplu, *hplu, *when;

   if (mode == 1) {
      s = src;
      od = src;
      while (*s) {
         if (*s == '%' && *(s+1) == '\"') {
	    src = s+2;
	    break;
	 }
	 s++;
      }
      if (src == od && t == Run_t) {
         *dst = 0;
	 return 0;
      }
      if (*src == '%' && *(src+1) == '\"') {
         *dst = 0;
	 return 0;
      }
      if (tim != -1) {
         sprintf(dst, "%02d:%02d ", (tim / 60), (tim % 60));
	 dst += strlen(dst);
      }
   }

   if (tim == -1) tim = curtim;
   tdiff = tim - curtim;
   adiff = ABS(tdiff);
   mdiff = adiff % 60;
   hdiff = adiff / 60;
   mplu = (mdiff == 1 ? "" : "s");
   hplu = (hdiff == 1 ? "" : "s");
   when = (tdiff < 0 ? "ago" : "from now");

   h = tim / 60;
   min = tim % 60;

   pm = (h < 12) ? "am" : "pm";
   hh = (h == 12) ? 12 : h % 12;

   ch = curtim / 60;
   cmin = curtim % 60;

   cpm = (ch < 12) ? "am" : "pm";
   chh = (ch == 12) ? 12 : ch % 12;

   *dst = 0;

   switch(d) {
      case 1:
      case 21:
      case 31: plu = "st"; break;

      case 2:
      case 22: plu = "nd"; break;

      case 3:
      case 23: plu = "rd"; break;

      default: plu = "th"; break;
   }


   while (c = *src++) {
     if (c == '\n') continue;
     else if (c != '%') { *dst++ = c; *dst = 0; }
     else {
         od = dst;
         c = *src++;
	 done = 0;
         if (diff <= 1) {
            switch(upper(c)) {
               case 'A':
               case 'B':
	       case 'C':
	       case 'E':
	       case 'F':
	       case 'G':
	       case 'H':
	       case 'I':
	       case 'J':
	       case 'K':
	       case 'L':
	       case 'U':
	       case 'V': sprintf(dst, "%s", (diff ? TOMORROW : TODAY));
		         dst += strlen(dst);
		         done = 1;
 		         break;

               default: done = 0;
            }
	 }

	 if (!done) switch(upper(c)) {
	    case 0: *dst = 0; return 0;

	    case '\"': if (mode == 1) {
		          *dst = 0;
			  return 0;
		       }
		       break;

   	    case 'A':
               sprintf(dst, "on %s, %d %s, %d", DayName[wkday], d,
		              MonthName[m], y);
               dst += strlen(dst);
	       break;

	    case 'B':
               sprintf(dst, "in %d days' time", diff);
	       dst += strlen(dst);
               break;

	    case 'C':
               sprintf(dst, "on %s", DayName[wkday]);
               dst += strlen(dst);
	       break;

	    case 'D':
	       sprintf(dst, "%d", d);
	       dst += strlen(dst);
               break;

	    case 'E':
	       sprintf(dst, "on %02d/%02d/%04d", d, m+1, y);
	       dst += strlen(dst);
               break;

	    case 'F':
	       sprintf(dst, "on %02d/%02d/%04d", m+1, d, y);
	       dst += strlen(dst);
               break;

	    case 'G':
               sprintf(dst, "on %s, %d %s", DayName[wkday], d, MonthName[m]);
	       dst += strlen(dst);
               break;

	    case 'H':
               sprintf(dst, "on %02d/%02d", d, m+1);
	       dst += strlen(dst);
               break;

	    case 'I':
               sprintf(dst, "on %02d/%02d", m+1, d);
	       dst += strlen(dst);
               break;

	    case 'J':
               sprintf(dst, "on %s, %s %d%s, %d", DayName[wkday],
		              MonthName[m], d, plu, y);
               dst += strlen(dst);
               break;

	    case 'K':
	       sprintf(dst, "on %s, %s %d%s", DayName[wkday],
	                      MonthName[m], d, plu);
	       dst += strlen(dst);
               break;

            case 'L':
	       sprintf(dst, "on %04d/%02d/%02d", y, m+1, d);
	       dst += strlen(dst);
               break;

	    case 'M':
	       sprintf(dst, "%s", MonthName[m]);
	       dst += strlen(dst);
               break;

	    case 'N':
	       sprintf(dst, "%d", m+1);
	       dst += strlen(dst);
               break;

	    case 'O':
               if (RealToday == JulianToday) sprintf(dst, " (today)");
               dst += strlen(dst);
               break;

	    case 'P':
	       sprintf(dst, (diff == 1 ? "" : "s"));
	       dst += strlen(dst);
               break;

	    case 'Q':
	       sprintf(dst, (diff == 1 ? "'s" : "s'"));
	       dst += strlen(dst);
               break;

	    case 'R':
	       sprintf(dst, "%02d", d);
	       dst += strlen(dst);
               break;

	    case 'S':
	       sprintf(dst, plu);
	       dst += strlen(dst);
               break;

	    case 'T':
	       sprintf(dst, "%02d", m+1);
	       dst += strlen(dst);
               break;

	    case 'U':
	       sprintf(dst, "on %s, %d%s %s, %d", DayName[wkday], d,
	                      plu, MonthName[m], y);
	       dst += strlen(dst);
               break;

	    case 'V':
	       sprintf(dst, "on %s, %d%s %s", DayName[wkday], d, plu,
	                      MonthName[m]);
	       dst += strlen(dst);
               break;

	    case 'W':
	       sprintf(dst, DayName[wkday]);
	       dst += strlen(dst);
               break;

	    case 'X':
	       sprintf(dst, "%d", diff);
	       dst += strlen(dst);
               break;

	    case 'Y':
	       sprintf(dst, "%d", y);
	       dst += strlen(dst);
               break;

	    case 'Z':
	       sprintf(dst, "%d", y % 100);
	       dst += strlen(dst);
               break;

	    case '1':
               if (tdiff == 0)
                  sprintf(dst, "now");
               else if (hdiff == 0)
	          sprintf(dst, "%d minute%s %s", mdiff, mplu, when);
	       else if (mdiff == 0)
	          sprintf(dst, "%d hour%s %s", hdiff, hplu, when);
               else
	          sprintf(dst, "%d hour%s and %d minute%s %s", hdiff, hplu, mdiff, mplu, when);
	       dst += strlen(dst);
               break;

            case '2':
               sprintf(dst, "at %d:%02d%s", hh, min, pm);
	       dst += strlen(dst);
	       break;

	    case '3': sprintf(dst, "at %02d:%02d", h, min);
	       dst += strlen(dst);
	       break;

	    case '4': sprintf(dst, "%d", tdiff);
	       dst += strlen(dst);
	       break;

	    case '5': sprintf(dst, "%d", adiff);
	       dst += strlen(dst);
	       break;

            case '6': sprintf(dst, when);
	       dst += strlen(dst);
	       break;

	    case '7': sprintf(dst, "%d", hdiff);
	       dst += strlen(dst);
	       break;

	    case '8': sprintf(dst, "%d", mdiff);
	       dst += strlen(dst);
	       break;

	    case '9': sprintf(dst, mplu);
	       dst += strlen(dst);
	       break;

	    case '0': sprintf(dst, hplu);
	       dst += strlen(dst);
	       break;

            case '!': sprintf(dst, (tdiff >= 0 ? "is" : "was"));
	       dst += strlen(dst);
	       break;

	    case '@': sprintf(dst, "%d:%02d%s", chh, cmin, cpm);
	       dst += strlen(dst);
	       break;

	    case '#': sprintf(dst, "%02d:%02d", ch, cmin);
	       dst += strlen(dst);
	       break;

#ifndef NL
#define NL "\n"
#endif
	    case '_': sprintf(dst, "%s", NL);
	       dst += strlen(dst);
	       break;

            default:
	       *dst++ = c;
	       *dst = 0;
	 }
	 if (isupper(c)) *od = upper(*od);
      }
   }
   if (t == Msg_t && mode == 0) *dst++ = '\n';
   *dst = 0;
   return 0;
}
