/*
 * SENDMSG: Utility to send messages for *.MSG or *.SQ? bases
 *
 * Created: 05/Dec/92
 * Updated: 05/Dec/92
 *
 * Written by Pete Kvitek of JV Dialogue 1st BBS (2:5020/6) +7-095-329-2192
 * Copyright (c) 1991 by JV DIALOGUE. All rights reserved.
 *
 */

#include <io.h>
#include <time.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "msgapi.h"

#ifdef __OS2__
#define MAXPATH   80
#define MAXDRIVE  3
#define MAXDIR    66
#define MAXFILE   9
#define MAXEXT    5
#define INCL_DOSFILEMGR
#include <os2.h>
#include <direct.h>
#else
#include <dir.h>
#endif

/*

 needed OS/2 types of Borland/DOS functions found in Kvitek's
 original sources Tue 06-20-1995 7:00:46 pm

*/
#ifdef __OS2__
#define chdir _chdir       /* that was easy :) */
void fnmerge(char *path, char *drive, char *dir,
            char *fname, char *ext);
void fnsplit(char *path, char *drive, char *dir, char *fname,
            char *ext);
int getdisk(void); /* get current default drive number */
int setdisk(int driveno); /* set default drive (0=A:, 1=B:, 2=C:) */
int getcurdir(ULONG drive, char *directory); /* drive 0=default, 1=A:,
                                                2=B: etc.) */

void fnmerge(char *path, char *drive, char *dir,
            char *fname, char *ext)
{
    _makepath(path, drive, dir, fname, ext);

}

void fnsplit(char *path, char *drive, char *dir, char *fname,
            char *ext)
{
    _splitpath(path, drive, dir, fname, ext);

}

getdisk()
{
    return (_getdrive() - 1);
}

setdisk(int driveno)
{
   int rc;
   rc = DosSetDefaultDisk( driveno + 1 );
   return rc;
}

getcurdir(ULONG drive, char *directory)
{
    APIRET rc;
    ULONG len = 126;

    ULONG gotDrive;
    ULONG LogDirMap;

    if(drive == 0){
    DosQueryCurrentDisk(&gotDrive, &LogDirMap);
    drive = gotDrive;
    }
    rc = DosQueryCurrentDir(drive, directory,
         &len);
    return rc;
}

/*

 end of OS/2 defs Tue 06-20-1995 8:31:32 pm AWG

*/
#endif  /* #ifdef __OS2__ */


/* //////////////////////////////////////////////////////////////////////////
// M o d u l e   d e c l a r a t i o n s                                   //
///////////////////////////////////////////////////////////////////////////*/

 /* Some useful defines */

 #define loop while (1)				/* Endless loop 'till break */
 #define numbof(a)  (sizeof(a)/sizeof(a[0]))	/* Number of elements */
 #define lengof(s)  (sizeof(s) - 1)		/* Length of the string */

 #define TRUE	     1
 #define FALSE	     0
 #define BOOL	     int

 /* Miscellaneous defines */

 #define VERSION  	"1.00"       		 /* Revision level */
 #ifdef __OS2__
 #define SMSG_NAME	"SendMsg/2"			 /* From/to name */
 #else
 #define SMSG_NAME  "SendMsg"
 #endif
 #define SMSG_ORIGIN    "\r--- "SMSG_NAME" v"VERSION"\r" /* Message origin */

 #define MSGTEXT	stdin			/* Message text input handle */
 #define AUXTEXT	stdout			/* Auxiliary text output handle */

 #define ISOPTION(ch)  ((ch)=='-'||(ch)=='/')	/* Command line option prefix */

 /* Program control flags in 'fsFlags' */

 #define FL_SQUISHMAIL	0x0001			/* Squish mail folder */

 /* Module variables */

 static char achMailPath[MAXPATH];		/* Path to mail folder */
 static unsigned fsFlags;			/* Control flags FL_ */
 static MSG * hMail;				/* Mail folder handle */
 static XMSG msg;				/* Message header info */

 /* Message attribute table */

 static struct {
   dword attr;		/* Attribute bit */
   char * psz;		/* Attribute name string */
   char ch;		/* Command line option char or null if none */
 } aMsgAttr[] = {
   MSGPRIVATE,		"Pvt",		'P',
   MSGCRASH,		"Crash",	'C',
   MSGREAD,		"Recv",		 0,
   MSGSENT,		"Sent",		 0,
   MSGFILE,		"File",		'F',
   MSGFWD,		"Transit",	 0,
   MSGORPHAN,		"Orphan",	 0,
   MSGKILL,		"Kill",		'K',
   MSGLOCAL,		"Local",	 0,
   MSGHOLD,		"Hold",		'H',
   MSGXX2,		"Rsvd2",	 0,
   MSGFRQ,		"Frq",		'Q',
   MSGRRQ,		"Rrq",		'R',
   MSGCPT,		"Cpt",		 0,
   MSGARQ,		"Arq",		 0,
   MSGURQ,		"Urq",		'U',
   MSGSCANNED,		"Scn",		 0,
 };

/* //////////////////////////////////////////////////////////////////////////
// M i s c e l l a n e o u s   s u b r o u t i n e s                       //
////////////////////////////////////////////////////////////////////////// */

/*
 * This subroutine displays logo
 */

 static void DoShowLogo(void)
 {
   fprintf(AUXTEXT,
     "\n"
     "Send Message v"VERSION", "__DATE__", "__TIME__"\n"
     "by Pete Kvitek of JV Dialogue 1st BBS, 2:5020/6\n"
     "Copyright (C) 1991 by JV Dialogue. All rights reserved.\n"
   );
 }

/*
 * This subroutine displays help
 */

 static void DoShowHelp(void)
 {
   fprintf(AUXTEXT,
     "Usage: SENDMSG <folder> <fromname,addr> [toname,addr] [-a<attr>] [-s<subj>]\n"
     "\n"
     "       folder    - specifies the mail folder path. If preceded\n"
     "                   with '$' then the folder is *.SQ? type base,\n"
     "                   otherwise it's assumed to be *.MSG style base\n"
     "       name,addr - specifies message from/to name and address\n"
     "       attr      - specifies message attributes: P,K,H,C,F,etc...\n"
     "       subj      - specifies message subject\n"
     "\n"
     "If any parameter includes spaces it should be enclosed in double quotes.\n"
     "The message body text file is supposed to be available through the standard\n"
     "input device.  Use redirection or piping characters to set it in.\n"
     "\n"
     "Example: SENDMSG c:\\mail Pete,2:5020/6 220/501 -aPL \"-sNew Files\" < NEWFILES.LST\n"
     "         echo Hi! | SENDMSG c:\\mail 2:5020/6 1.1 -aPLF \"-sc:\\autoexec.bat\"\n"
   );
 }

/*
 * This sub routine builds a fully qualified path
 */

 BOOL BuildFullPath(char * pszDest, char * pszSrc)
 {
   char achDrive[MAXDRIVE];
   char achDir[MAXDIR];
   char achFile[MAXFILE];
   char achExt[MAXEXT];
   char achCurDir[MAXDIR];
   int iCurDrive;

   /* Decompose supplied path */

   fnsplit(pszSrc, achDrive, achDir, achFile, achExt);

   /* Preserve current drive */

   iCurDrive = getdisk();

   /* Check if drive specified in the supplied path and if not,
   // assume the current one */

   if (!achDrive[0]) {
     strcpy(achDrive, "A:");
     achDrive[0]+= (char) iCurDrive;
   }
   /* Set the current drive to the requested one and check if ok.
   // If failed, restore original drive and return error */


   setdisk(toupper(achDrive[0]) - 'A');
   if (getdisk() != toupper(achDrive[0]) - 'A') {
     setdisk(iCurDrive);
     return FALSE;
   }



   /* Preserve the current directory on the reqested drive and
   // check if ok, otherwise restore initial current drive and return */

   strcpy (achCurDir, "\\");


   if (getcurdir(0, &achCurDir[1])) {
     setdisk(iCurDrive);
     return FALSE;
   }

   /* Check if directory specified and make it current */

   if (achDir[0]) {

     /* Kill trailing back slash if it's not the only character
     // of the directory specification */

     if ((achDir[strlen(achDir) - 1] == '\\') && (strlen(achDir) > 1 ))
       achDir[strlen(achDir) - 1] = '\0';

     /* Change to the specified directory and check if ok. If failed,
     // restore the inital directory on the requested drive and change
     // to the initial drive */

     if (chdir(achDir)) {
       chdir(achCurDir);
       setdisk(iCurDrive);
       return FALSE;
     }
   }

   /* So we managed to make a requested directory current on the
   // requested drive. Now get its full specification and this
   // will be what we're after. If failed, just restore things back
   */
   strcpy(achDir, "\\");
   if (getcurdir(0, &achDir[1])) {
     chdir(achCurDir);
     setdisk(iCurDrive);
     return FALSE;
   }

   /* Compose the fully qualified file name and restore
   // the inital directory on the requested drive and change
   // to the initial drive
   */
   fnmerge(pszDest, achDrive, achDir, achFile, achExt);
   chdir(achCurDir);
   setdisk(iCurDrive);

   return TRUE;
 }

/*
 * This routine scans in z:n/n.p address specification
 */

 static char * DoScanNetAddr(NETADDR * pnetAddr, char * psz)
 {
   char * pch, * pchEnd, * pchNext, ch;

   /* Skip through the leading spaces and fix up the end */

   for (pch = psz; isspace(*pch); pch++);
   for (pchNext = pch; *pchNext && !isspace(*pchNext); pchNext++);
   ch = *pchNext; *pchNext = '\0'; pchEnd = pch;

   /* Scan in the zone if any */

   if (*pch && !isspace(*pch) && strchr(pch, ':')) {
     while (isdigit(*pchEnd)) pchEnd++;
     if (*pchEnd != ':' || pch == pchEnd) return NULL;
     pnetAddr->zone = atoi(pch);
     pch = ++pchEnd;
     if (!isdigit(*pch) || !strchr(pch, '/')) return NULL;
   }

   /* Scan in the net if any */

   if (*pch && !isspace(*pch) && strchr(pch, '/')) {
     while (isdigit(*pchEnd)) pchEnd++;
     if (*pchEnd != '/' || pch == pchEnd) return NULL;
     pnetAddr->net = atoi(pch);
     pch = ++pchEnd;
     if (!isdigit(*pch)) return NULL;
   }

   /* Scan in the node if any */

   if (*pch && !isspace(*pch) && *pch != '.') {
     while (isdigit(*pchEnd)) pchEnd++;
     if (*pchEnd != '.' && !isspace(*pchEnd) && *pchEnd) return NULL;
     pnetAddr->node = atoi(pch);
     pch = pchEnd;
   }

   /* Scan in the point if any */

   if (*pch != '.') {
     if (!isspace(*pch) && *pch) return NULL;
     pnetAddr->point = 0;
   } else {
     for (pchEnd = ++pch; isdigit(*pchEnd); pchEnd++);
     if (!isspace(*pchEnd) && *pchEnd) return NULL;
     pnetAddr->point = atoi(pch);
     pch = pchEnd;
   }

   /* Restore the zeroed trailing characters */

   *pchNext = ch;

   /* Check if zone or net is zero and return */

   return (pnetAddr->zone && pnetAddr->net) ? pchNext : NULL;
 }

/*
 * This subroutine scans name and network address
 */

 static BOOL DoScanNameAddr(char * psz, NETADDR * pnetAddr,
                            char * pszName, short cchName)
 {
   char * pch;
   short cch;

   /* Check if there is a name and scan it in */

   if (isdigit(*psz))
     pch = psz;
   else
     if ((pch = strchr(psz, ',')) == NULL) {
       return FALSE;
     } else {
       cch = min((short)(pch - psz), cchName);
       memcpy(pszName, psz, cch);
       pszName[cch] = '\0';
       pch++;
     }

   /* Scan in the network address if any */

   DoScanNetAddr(pnetAddr, pch);

   return TRUE;
 }

/*
 * This subroutine scans message attribute specification
 */

 static void DoScanMsgAttr(char * psz, dword * pattr)
 {
   short iAttr;

   /* Scan through all the specified message attributes */

   for (; *psz; psz++) {
     for (iAttr = 0; toupper(*psz) != aMsgAttr[iAttr].ch; iAttr++)
       if (iAttr >= numbof(aMsgAttr)) {
	 fprintf(AUXTEXT, "Unknown message attribute: '%s'\n", psz);
	 exit(EXIT_FAILURE);
       }
     *pattr|= aMsgAttr[iAttr].attr;
   }
 }

/*
 * This subroutine prints out a message header
 */

 static void DoPrintMsgHeader(XMSG * pMsg)
 {
   int iAttr;
   /* Print out the message header */

   fprintf(AUXTEXT,
	   "  To: %s, %u:%u/%u.%u\n"
	   "From: %s, %u:%u/%u.%u\n"
	   "Subj: %s\n",
	   pMsg->to,   pMsg->dest.zone, pMsg->dest.net, pMsg->dest.node, pMsg->dest.point,
	   pMsg->from, pMsg->orig.zone, pMsg->orig.net, pMsg->orig.node, pMsg->orig.point,
	   pMsg->subj
	  );

   /* Print out the message attribute if any */

   if (pMsg->attr) {
     fprintf(AUXTEXT, "Attr:");
     for (iAttr = 0; iAttr < numbof(aMsgAttr); iAttr++)
       if (pMsg->attr & aMsgAttr[iAttr].attr)
	 fprintf(AUXTEXT, " %s", aMsgAttr[iAttr].psz);
     putc('\n', AUXTEXT);
   }
 }

/*
 * This subroutine to process the command line parameters
 */

 static void DoProcCmdLine(int cArg, char * apszArg[])
 {
   char * psz;
   short iArg = 1;

   /* Display the logo screen and check if there are any
   // command line parameters specified and if not, display
   // help screen and quit now...
   */

   DoShowLogo();
   if (cArg == 1) {
     DoShowHelp();
     exit(EXIT_FAILURE);
   }

   /* Scan in the mail folder path specification and check if it's
   // a *.SQ? type database
   */
   if (apszArg[iArg][0] == '$') {
     psz = &apszArg[iArg][1]; fsFlags|= FL_SQUISHMAIL;
   } else {
     psz = &apszArg[iArg][0];
   }

   /* Build the fully qualified path and make it upper case */

   if (!BuildFullPath(achMailPath, psz)) {
     fprintf(AUXTEXT, "Invalid mail folder path: '%s'\n", psz);
     exit(EXIT_FAILURE);
   } else {
     strupr(achMailPath);
     iArg++;
   }

   /* Scan in the 'From' name/address */

   if (iArg >= cArg) {
     fprintf(AUXTEXT, "Missing 'From' address/name\n");
     exit(EXIT_FAILURE);
   } else
     if (!DoScanNameAddr(apszArg[iArg], &msg.orig, msg.from, lengof(msg.from)) ||
	  msg.orig.zone == 0 || msg.orig.net == 0) {
       fprintf(AUXTEXT, "Invalid 'From' address: '%s'\n", apszArg[iArg]);
       exit(EXIT_FAILURE);
     } else {
       memcpy(&msg.dest, &msg.orig, sizeof(msg.dest));
       iArg++;
     }

   /* Scan in the 'To' name/address if any */

   if (iArg >= cArg || ISOPTION(apszArg[iArg][0])) {
     memcpy(&msg.dest, &msg.orig, sizeof(msg.dest));
   } else
     if (!DoScanNameAddr(apszArg[iArg], &msg.dest, msg.to, lengof(msg.to)) ||
	  msg.dest.zone == 0 || msg.dest.net == 0) {
       fprintf(AUXTEXT, "Invalid 'To' address: '%s'\n", apszArg[iArg]);
       exit(EXIT_FAILURE);
     } else
       iArg++;

   /* Process the command line options if any */

   for (; iArg < cArg; iArg++)
     if (!ISOPTION(apszArg[iArg][0])) {
       fprintf(AUXTEXT, "Invalid option: '%s'\n", apszArg[iArg]);
       exit(EXIT_FAILURE);
     } else
       switch (tolower(apszArg[iArg][1])) {
	 case 's': /* Message subject */
		   strncpy(msg.subj, &apszArg[iArg][2], lengof(msg.subj));
		   break;
	 case 'a': /* Message attributes */
		   DoScanMsgAttr(&apszArg[iArg][2], &msg.attr);
		   break;
	 default:  fprintf(AUXTEXT, "Unknown option: '%s'\n", apszArg[iArg]);
		   exit(EXIT_FAILURE);
       }

   /* Set up default message header strings if not defined yet */

   if (!msg.to[0]) strcpy(msg.to, msg.attr & MSGPRIVATE ? "SysOp" : "All");
   if (!msg.from[0]) strcpy(msg.from, "SysOp");
   if (!msg.subj[0]) strcpy(msg.subj, "<none>");

   /* Explicitely set the 'Local' message attribute */

   msg.attr|= MSGLOCAL;
 }

/*
 * This subroutine returns an API error string
 */

 static char * DoGetApiError(void)
 {
   switch (msgapierr) {
     case MERR_NOMEM: return "not enough memory";
     case MERR_NOENT: return "area does not exist";
     case MERR_BADF:  return "area is damaged";
     default:         return "error unknown";
   }
 }

/*
 * This subroutine opens a mail folder
 */

 static BOOL DoOpenMailFolder(void)
 {
   struct _minf minf;

   /* Initialize Scott's API and check if ok */

   minf.req_version = 0;	/* MSGAPI_VERSION; !! not defined !! */
   minf.def_zone = msg.orig.zone;

   if (MsgOpenApi(&minf) == -1) {
     fprintf(AUXTEXT, "MsgAPI initialization failed\n");
     exit(EXIT_FAILURE);
   }

   /* Open the mail folder, lock it and check if ok */

   if ((hMail = MsgOpenArea( achMailPath
			   , MSGAREA_NORMAL
			   , fsFlags & FL_SQUISHMAIL ? MSGTYPE_SQUISH : MSGTYPE_SDM
			   )) == NULL || MsgLock(hMail) == -1) {
     fprintf(AUXTEXT, "Can't open folder %s%s -- %s\n",
	     achMailPath, fsFlags & FL_SQUISHMAIL ? ".SQ?" : "\\*.MSG",
	     DoGetApiError());
     exit(EXIT_FAILURE);
   }

   return TRUE;
 }

/*
 * This subroutine closes a mail folder
 */

 static BOOL DoCloseMailFolder(void)
 {
   /* Check if opened */

   if (hMail == NULL)
     return FALSE;

   /* Unlock and close the netmail folder and check if ok */

   if (MsgUnlock(hMail) == -1 || MsgCloseArea(hMail) == -1) {
     fprintf(AUXTEXT, "Can't close mail folder: %s\n", DoGetApiError());
     exit(EXIT_FAILURE);
   }

   /* Shut down the Scott's API */

   if (MsgCloseApi() == -1) {
     fprintf(AUXTEXT, "MsgAPI shutdown failed\n");
     exit(EXIT_FAILURE);
   }

   return TRUE;
 }

/*
 * This subroutine sets message creation date/stamp
 */

 static void DoSetMsgDateTime(XMSG * pMsg)
 {
   time_t sec;
   struct tm * pTime;

   /* Get the current time */

   sec = time(NULL); pTime = localtime(&sec);

   /* Set message date time stamp */

   pMsg->date_written.date.yr = pTime->tm_year - 80;
   pMsg->date_written.date.mo = pTime->tm_mon + 1;
   pMsg->date_written.date.da = pTime->tm_mday;

   pMsg->date_written.time.hh = pTime->tm_hour;
   pMsg->date_written.time.mm = pTime->tm_min;
   pMsg->date_written.time.ss = pTime->tm_sec;
 }

/*
 * This subroutine creates a new message
 */

 static void DoCreateMessage(XMSG * pMsg)
 {
   short cch, ich;
   long cchTotal;
   char ach[512];
   char * pch;
   MSGH * hMsg;

   /* Calculate length of the message body. Note that the
   // origin line will not be written for the empty messages
   */
   if ((cchTotal = filelength(fileno(MSGTEXT))) > 0)
     cchTotal+= sizeof(SMSG_ORIGIN);

   /* Open the new message and check if ok */

   if ((hMsg = MsgOpenMsg(hMail, MOPEN_CREATE, 0)) == NULL) {
     fprintf(AUXTEXT, "Can't create message\n");
     exit(EXIT_FAILURE);
   }

   /* Set the message creation date/time stamp */

   DoSetMsgDateTime(pMsg);

   /* Write out the new message header and check if ok */

   if (MsgWriteMsg(hMsg, FALSE, pMsg, NULL, 0L, cchTotal, 0L, NULL) == -1) {
     fprintf(AUXTEXT, "Can't write message header\n");
     MsgCloseMsg(hMsg);
     exit(EXIT_FAILURE);
   }

   /* Check if we have to write message body */

   if (cchTotal) {

     /* Write out new message text if any. Replace all \n with \r
     // in order to conform FidoNet(tm) message body standards
	 */
     while ((cch = fread(ach, 1, sizeof(ach), MSGTEXT)) > 0) {
       for (pch = ach, ich = 0; ich < cch; pch++, ich++)
	 if (*pch == '\n') *pch = '\r';
       if (MsgWriteMsg(hMsg, TRUE, NULL, ach, cch, 0L, 0L, NULL) == -1) {
	 fprintf(AUXTEXT, "Can't write message body\n");
	 MsgCloseMsg(hMsg);
	 exit(EXIT_FAILURE);
       }
     }

     /* Write out the origin line. Note that the trailing null is also
     // written since we're using 'sizeof()'
	 */
     if (MsgWriteMsg(hMsg, TRUE, NULL, SMSG_ORIGIN, sizeof(SMSG_ORIGIN),
		     0L, 0L, NULL) == -1) {
       fprintf(AUXTEXT, "Can't write origin line\n");
       MsgCloseMsg(hMsg);
       exit(EXIT_FAILURE);
     }
   }

   /* Close the message just created and display the results */

   MsgCloseMsg(hMsg);

   DoPrintMsgHeader(&msg);

   fprintf(AUXTEXT, "Sent: %lu byte(s)\n"
		    "Mail: %s%s\n",
		     cchTotal, achMailPath,
		     fsFlags & FL_SQUISHMAIL ? ".SQ?" : "\\*.MSG");
 }

/* //////////////////////////////////////////////////////////////////////////
// M a i n . . .                                                           //
////////////////////////////////////////////////////////////////////////// */

 void main(int argc, char * argv[])
 {
   DoProcCmdLine(argc, argv);
   DoOpenMailFolder();
   DoCreateMessage(&msg);
   DoCloseMailFolder();
   exit(EXIT_SUCCESS);
 }

/*
 * End of SENDMSG.C
 */
